
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '1.25.3',
    'branch_nick': 'master',
    'revision_id': '612972bad9dba7b44593b8d0aae8e5c5a0b2d515',
    'revno': 10573
}
